unit MainFormU;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls,
  System.Generics.Collections, System.Net.URLClient, System.Net.HttpClient, System.Net.HttpClientComponent;

type
  TMainForm = class(TForm)
    btnTStrings: TButton;
    MemoLog: TMemo;
    btnStreams: TButton;
    btnTArray: TButton;
    btnTextReader: TButton;
    btnDirectoryEnum: TButton;
    btnCompEnumerator: TButton;
    btnTPath: TButton;
    btnHash: TButton;
    btnSquids: TButton;
    btnThreadPool: TButton;
    btnCurl: TButton;
    NetHTTPClient1: TNetHTTPClient;
    NetHTTPClient: TButton;
    btnFVL: TButton;
    procedure btnTStringsClick(Sender: TObject);
    procedure btnStreamsClick(Sender: TObject);
    procedure btnTArrayClick(Sender: TObject);
    procedure btnTextReaderClick(Sender: TObject);
    procedure btnDirectoryEnumClick(Sender: TObject);
    procedure btnCompEnumeratorClick(Sender: TObject);
    procedure btnTPathClick(Sender: TObject);
    procedure btnHashClick(Sender: TObject);
    procedure btnSquidsClick(Sender: TObject);
    procedure btnThreadPoolClick(Sender: TObject);
    procedure btnCurlClick(Sender: TObject);
    procedure NetHTTPClientClick(Sender: TObject);
    procedure NetHTTPClient1ReceiveDataEx(const Sender: TObject; AContentLength, AReadCount: Int64; AChunk: Pointer;
      AChunkLength: Cardinal; var AAbort: Boolean);
    procedure NetHTTPClient1ValidateServerCertificate(const Sender: TObject; const ARequest: TURLRequest;
      const Certificate: TCertificate; var Accepted: Boolean);
    procedure btnFVLClick(Sender: TObject);
  private
    procedure LogClear;
    procedure OnValidateCertificateEvent(const Sender: TObject; const ARequest: TURLRequest; const Certificate: TCertificate; var Accepted: Boolean);
    procedure Log(const Value: String); overload;
    procedure Log(const Value: String; const BoolValue: Boolean); overload;
    procedure Log(const Value: String; const StrValue: String); overload;
    procedure Log(const Value: String; const IntValue: Integer); overload;
    procedure Log(const Value: String; const Int64Value: UInt64); overload;
    procedure Log(const Value: String; const IntValues: TArray<Integer>); overload;
    procedure PrintSet(const VarName: String; const HashSet: THashSet<String>);
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

uses
  System.IOUtils, System.Generics.Defaults, System.NetEncoding.Sqids,
  System.Hash, System.Threading, System.SyncObjs, System.UIConsts,
  System.Net.HttpClient.Curl {required to use curl}
  ;

{$R *.dfm}

var
  gInc: Int64 = 0;

type
  TMyClass = class
  private
    fValue: String;
  public
    constructor Create(Value: String);
    function ToString: string; override;
  end;

procedure TMainForm.btnTArrayClick(Sender: TObject);
begin
  LogClear;
  Log('** TArray - Helper Methods **');
  var lMyArr := ['daniele','teti','rome','italy','daniele','teti'];

  Log('TArray.IndexOf<string>(lMyArr, ''rome'')', TArray.IndexOf<string>(lMyArr, 'rome'));
  Log('TArray.LastIndexOf<string>(lMyArr, ''daniele'')', TArray.LastIndexOf<string>(lMyArr, 'daniele'));
  Log('TArray.Contains<string>(lMyArr, ''rome'')', TArray.Contains<string>(lMyArr, 'rome'));
  Log('TArray.ToString<string>(lMyArr, '','')', TArray.ToString<string>(lMyArr, ','));
  Log('TArray.ToString<string>(lMyArr, '','', ''['','']'')', TArray.ToString<string>(lMyArr, ',', '[',']'));


  var lMyObjArray: TArray<TMyClass> := [
                                          TMyClass.Create('Ciao'),
                                          TMyClass.Create('Daniele'),
                                          TMyClass.Create('Teti')
                                       ];

  //Uses TMyClass.ToString
  Log('TArray.ToString<TMyClass>(lMyObjArray, '','')', TArray.ToString<TMyClass>(lMyObjArray, ','));

  // Search for identity
  Log('TArray.Contains<TMyClass>(lMyObjArray, ''rome'')', TArray.Contains<TMyClass>(lMyObjArray, lMyObjArray[1]));

  // Search for custom criteria
  var lToFind := TMyClass.Create('Daniele');
  try
    Log('TArray.Contains<TMyClass>(lMyObjArray, ...new object..., ...criteria...)', TArray.Contains<TMyClass>(
      lMyObjArray,
      lToFind,
      TComparer<TMyClass>.Construct(
        function(const Left, Right: TMyClass): Integer
        begin
          //"Right" is "lToFind"
          Result := String.Compare(Left.fValue, Right.fValue);
        end)
      ));
  finally
    lToFind.Free;
  end;

  TArray.FreeValues<TMyClass>(lMyObjArray);
end;

procedure TMainForm.btnTextReaderClick(Sender: TObject);
begin
  LogClear;
  Log('** TStreamReader - TFile.OpenText **');
  var lFile := TFile.OpenText('MyFile.txt');
  try
    for var lLine in lFile do
      Log(lLine);
  finally
    lFile.Free;
  end;


  Log('** TStreamReader.Create(AnotherStream) **');
  var lSR := TStreamReader.Create(TFileStream.Create('MyFile.txt', fmOpenRead));
  try
    lSR.OwnStream;
    for var lLine in lSR do
      Log(lLine);
  finally
    lSR.Free;
  end;

  Log('** TFile.GetLinesEnumerator **');
  for var lLine in TFile.GetLinesEnumerator('MyFile.txt') do
    Log(lLine);
end;

procedure TMainForm.btnThreadPoolClick(Sender: TObject);
begin
  LogClear;
  Log('** TThreadPool **');
  var lValue := 0;
  var lPool := TThreadPool.Create;
  try
    lPool.Interactive := True;
    lPool.MaxWorkerThreads := 4;
    lPool.MinWorkerThreads := 4;
    lPool.OnThreadStart :=     procedure (Thread: TThread)
                               begin
                                 var lID := Thread.Current.ThreadID;
                                 TThread.Queue(nil, procedure begin Log('ThreadStart'.PadRight(20), lID); end);
                               end;
    lPool.OnThreadTerminate := procedure (Thread: TThread)
                               begin
                                 var lID := Thread.Current.ThreadID;
                                 TThread.Queue(nil, procedure begin Log('ThreadTerminate'.PadRight(20), lID); end);
                               end;

    var lTasks: TArray<ITask> := [];
    for var I := 1 to 50 do
    begin
      lTasks := lTasks + [TTask.Run(
        procedure
        begin
          var lID := TThread.Current.ThreadID;
          TThread.Queue(nil, procedure begin Log('Task'.PadRight(20), lID); end);
          for var I := 1 to 5 do
          begin
            Sleep(10);
            TInterlocked.Add(lValue, 1);
          end;
        end, lPool)];
    end;
    TTask.WaitForAll(lTasks);
  finally
    lPool.Free;
  end;
end;

procedure TMainForm.btnTStringsClick(Sender: TObject);
begin
  LogClear;
  Log('**TStrings**');
  var lSL := TStringList.Create;
  try
    lSL.Add('HELLO WORLD');
    lSL
      .AddPair('Nome','Daniele')
      .AddPair('Cognome','Teti')
      .AddPair('Altro','Foto', btnTStrings);

    Log('lSL.IsEmpty', lSL.IsEmpty);
    Log('lSL.Contains(''HELLO WORLD'')', lSL.Contains('HELLO WORLD'));
    Log('lSL.Contains(''Nome'')', lSL.Contains('Nome'));
    Log('lSL.ContainsName(''Nome'')', lSL.ContainsName('Nome'));
    Log('lSL.ContainsObject(Button1)', lSL.ContainsObject(btnTStrings));
    Log('lSL.ContainsObject(Self)', lSL.ContainsObject(Self));
  finally
    lSL.Free;
  end;
end;

procedure TMainForm.Log(const Value: String);
begin
  var lHeader := Value.StartsWith('**');
  if lHeader then
    MemoLog.Lines.Add(StringOfChar('-', Value.Length));
  MemoLog.Lines.Add(Value);
  if lHeader then
    MemoLog.Lines.Add(StringOfChar('-', Value.Length));
  MemoLog.Update;
end;

procedure TMainForm.btnCurlClick(Sender: TObject);
begin
  LogClear;
  Log('** Curl support **');
  var lClient := THTTPClient.Create;
  try
    lClient.OnValidateServerCertificate := OnValidateCertificateEvent;
    var lResp := lClient.Get('https://www.danieleteti.it');
    Log('Homepage size in bytes', lResp.ContentLength);
    Log('** RESPONSE **');
    Log(lResp.ContentAsString());
  finally
    lClient.Free;
  end;
end;

procedure TMainForm.btnCompEnumeratorClick(Sender: TObject);
begin
  // System.Classes.TComponentEnumerator
  LogClear;
  Log('** Components Enumerator (TComponent+) **');
  for var lComponent in Self do
  begin
    Log(Format('%s (%s)', [lComponent.Name, lComponent.ClassName]));
  end;


  // Vcl.Controls.TControlEnumerator
  Log('** Controls Enumerator (TWinControl+) ceftAll **');
  for var lControl in Self.GetControls([ceftAll]) do
  begin
    Log(Format('%s (%s)', [lControl.Name, lControl.ClassName]));
  end;

  Log('** Controls Enumerator (TWinControl+) **');
  for var lControl in Self.GetControls([ceftAll]) do
  begin
    Log(Format('%s (%s)', [lControl.Name, lControl.ClassName]));
  end;

end;

procedure TMainForm.btnTPathClick(Sender: TObject);
begin
  LogClear;
  Log('** TPath (new sys folders) **');
  Log('TPath.GetDesktopPath', TPath.GetDesktopPath); //ok on windows, "guess" on other platforms
  Log('TPath.GetAppPath', TPath.GetAppPath);

  Log('** TPath (more powerful combine) **');
  Log('TPath.Combine(...with 2 params...)', TPath.Combine(TPath.GetAppPath, 'path1'));
  Log('TPath.Combine(...with 3 params...)', TPath.Combine(TPath.GetAppPath, 'path1', 'path2'));
  Log('TPath.Combine(...with 4 params...)', TPath.Combine(TPath.GetAppPath, 'path1', 'path2', 'path3'));
end;

procedure TMainForm.btnDirectoryEnumClick(Sender: TObject);
begin
  LogClear;

  Log('** TDirectory.GetFilesEnumerator **');
  for var lLine in TDirectory.GetFilesEnumerator('.') do
  begin
    Log(lLine);
  end;

  Log('** TDirectory.GetDirectoriesEnumerator **');
  for var lLine in TDirectory.GetDirectoriesEnumerator('..\..') do
  begin
    Log(lLine);
  end;

  Log('** TDirectory.GetFileSystemEntriesEnumerator **');
  for var lLine in TDirectory.GetFileSystemEntriesEnumerator('..\..') do
  begin
    Log(lLine);
  end;

  Log('** TDirectory.GetFileSystemEntriesEnumerator (with filters) **');
  for var lLine in TDirectory.GetFileSystemEntriesEnumerator('..\..',
                              function(const Path: string; const SearchRec: TSearchRec): Boolean
                              begin
                                // I want directories and files bigger than 10 bytes
                                Result := (SearchRec.Attr = faDirectory) or (SearchRec.Size > 1024);
                              end) do
  begin
    Log(lLine);
  end;
end;

procedure TMainForm.btnFVLClick(Sender: TObject);
begin
  LogClear;
  Log('** THashFNV1a32/64 **');
  Log('THashFNV1a64.GetHashString(''Hello There'')', THashFNV1a64.GetHashString('Hello There'));
  Log('THashFNV1a32.GetHashString(''Hello There'')', THashFNV1a32.GetHashString('Hello There'));
  Log('THashFNV1a64.GetHashValue(''Hello There'')', UInt64(THashFNV1a64.GetHashValue('Hello There')));
end;

procedure TMainForm.btnHashClick(Sender: TObject);
begin
  LogClear;
  Log('** THashSet & TObjectHasSet **');
  var lSet1 := THashSet<String>.Create(['daniele', 'teti', 'hello', 'world', 'hello', 'daniele']);
  try
    Log('Check... a set cannot contains duplicates');
    PrintSet('lSet1', lSet1);

    lSet1.AddRange(['tic','toc']);
    lSet1.Remove('tic');
    Log('lSet1.GetOrAdd(''ping'')', lSet1.GetOrAdd('ping'));
    Log('lSet1.Contains(''daniele'')', lSet1.Contains('daniele'));
  finally
    lSet1.Free;
  end;

  Log('** Except **');
  lSet1 := THashSet<String>.Create(['daniele', 'teti', 'hello', 'world']);
  try
    var lSet2 := THashSet<String>.Create(['daniele','world','pippo']);
    PrintSet('lSet1', lSet1);
    PrintSet('lSet2', lSet2);
    lSet1.ExceptWith(lSet2);
    PrintSet('lSet1 - lSet2 {except}', lSet1);
  finally
    lSet1.Free;
  end;

  Log('** Intersection **');
  lSet1 := THashSet<String>.Create(['daniele', 'teti', 'hello', 'world', 'hello', 'daniele']);
  try
    var lSet2 := THashSet<String>.Create(['daniele','world','pippo']);
    PrintSet('lSet1', lSet1);
    PrintSet('lSet2', lSet2);
    lSet1.IntersectWith(lSet2); //intersezione
    PrintSet('lSet1 * lSet2 {intersection}', lSet1);
  finally
    lSet1.Free;
  end;

  Log('** Union **');
  lSet1 := THashSet<String>.Create(['daniele', 'teti', 'hello', 'world', 'hello', 'daniele']);
  try
    var lSet2 := THashSet<String>.Create(['daniele','world','pippo']);
    PrintSet('lSet1', lSet1);
    PrintSet('lSet2', lSet2);
    lSet1.UnionWith(lSet2); //union
    PrintSet('lSet1 + lSet2 {union}', lSet1);
  finally
    lSet1.Free;
  end;

  Log('** Overlaps **');
  lSet1 := THashSet<String>.Create(['daniele', 'teti', 'hello', 'world', 'hello', 'daniele']);
  try
    var lSet2 := THashSet<String>.Create(['daniele','world','pippo']);
    PrintSet('lSet1', lSet1);
    PrintSet('lSet2', lSet2);
    Log('lSet1.Overlaps(lSet2)', lSet1.Overlaps(lSet2)); //overlaps
  finally
    lSet1.Free;
  end;

  Log('** IsSubSet and IsSuperSet **');
  lSet1 := THashSet<String>.Create(['daniele', 'teti', 'hello', 'world', 'hello', 'daniele']);
  try
    var lSet2 := THashSet<String>.Create(['daniele','world']);
    PrintSet('lSet1', lSet1);
    PrintSet('lSet2', lSet2);
    Log('lSet1.IsSubsetOf(lSet2)', lSet1.IsSubsetOf(lSet2));
    Log('lSet1.IsSuperSetOf(lSet2)', lSet1.IsSuperSetOf(lSet2));
  finally
    lSet1.Free;
  end;

end;

procedure TMainForm.btnSquidsClick(Sender: TObject);
begin
  LogClear;
  Log('** Squids **');
  var lLength := 50;

  var lAlphabet := 'JXx6Ue7jLqBk9cZmwA2szgdYNRP4lhOErDvQF1GTVIM3fao5t8WuSb0HyipKnC';
  Log('lAlphabet', lAlphabet);
  Log('lLength', lLength);
  var lSqids := TSqidsEncoding.Create(lAlphabet, lLength);
  try
    var lEncoded := lSqids.Encode(1234); // The encoded value will be 'AWwcaNAXqs'
    Log('lSqids.Encode(1234)', lEncoded);

    var lDecodedSingle := lSqids.DecodeSingle(lEncoded); // The decoded value will be an integer with the value: 1234
    Log('lSqids.DecodeSingle(lEncoded)', lDecodedSingle);

    var lDecoded := lSqids.Decode(lEncoded); // The decoded value will be and array with one value: 1234

    var lEncodedMulti := lSqids.Encode([123,234,345]); // The encoded value will be 'l0cCVaQOQI'
    Log('lSqids.Encode([123,234,345])', lEncodedMulti);
    var lDecodedMulti := lSqids.Decode(lEncodedMulti); // [123,234,345]
    Log('lSqids.Decode(lEncodedMulti)', lDecodedMulti);
  finally
    lSqids.Free;
  end;


  Log('** Changing Alphabet and MinLength **');
  lSqids := TSqidsEncoding.Create('abc', 5);
  try
    var lEncoded := lSqids.Encode(1234);
    Log('lSqids.Encode(1234)', lEncoded);

    var lDecodedSingle := lSqids.DecodeSingle(lEncoded);
    Log('lSqids.DecodeSingle(lEncoded)', lDecodedSingle);
  finally
    lSqids.Free;
  end;
end;

procedure TMainForm.btnStreamsClick(Sender: TObject);
begin
  LogClear;
  Log('** TProxySubrangeStream **');
  var lFS1 := TFile.Open('file1.dat', TFileMode.fmOpenOrCreate);
  try
    var lSW := TStreamWriter.Create(lFS1);
    try
      lSW.Write('Hello Cruel World');
    finally
      lSW.Free;
    end;

    // Helper class for reading/writing a segment of another stream.
    //   For example, it is used by TZipFile to provide access to a "stored" file.
    var lProxySubRange := TProxySubrangeStream.Create(lFS1, 5, 7);
    try
      var lSR := TStreamReader.Create(lProxySubRange, False);
      try
        Log(lSR.ReadToEnd);
      finally
        lSR.Free;
      end;

      lProxySubRange.Position := 0;

      var lSW2 := TStreamWriter.Create(lProxySubRange);
      try
        lSW2.Write(' NICE ');
      finally
        lSW2.Free;
      end;
    finally
      lProxySubRange.Free;
    end;
  finally
    lFS1.Free;
  end;

  Log('** TProxyAggregateStream **');
  ///  Helper class for reading concatenation of another streams.
  ///    For example, it is used by TMultipartFormData to aggregate form elements
  ///    and represent them as a single stream.
  var lProxyAggregate := TProxyAggregateStream.Create; // it's read-only
  try
    var lFile1 := TFile.Create('file1.dat');
    try
      for var I := 0 to 4 do // 20 bytes
        lFile1.WriteData(I);
      var lMem2 := TMemoryStream.Create;
      try
        for var I := 5 to 9 do // 20 bytes
          lMem2.WriteData(I);
        lProxyAggregate.AddStream(lFile1);
        lProxyAggregate.AddStream(lMem2);
        //lProxyAggregate.AddStream(lMem2); {you can also replicate}
        lProxyAggregate.Position := 0;

        Log('lProxyAggregate.Size', lProxyAggregate.Size.ToString);

        var lData: Integer := 0;
        for var I := 0 to lProxyAggregate.Size - 1 do
        begin
          lProxyAggregate.ReadData(lData);
          Log('Byte[' + I.ToString + '] = ' + lData.ToString);
        end;
      finally
        lMem2.Free;
      end;
    finally
      lFile1.Free;
    end;
  finally
    lProxyAggregate.Free;
  end;
end;

procedure TMainForm.Log(const Value, StrValue: String);
begin
  Log(Value + ' = ' + StrValue);
end;

procedure TMainForm.Log(const Value: String; const IntValue: Integer);
begin
  Log(Value + ' = ' + IntValue.ToString);
end;

procedure TMainForm.LogClear;
begin
  MemoLog.Clear;
end;

procedure TMainForm.NetHTTPClient1ReceiveDataEx(
  const Sender: TObject;
  AContentLength, AReadCount: Int64;
  AChunk: Pointer;
  AChunkLength: Cardinal; var AAbort: Boolean);
begin
  Log('NetHTTPClient1ReceiveDataEx', AReadCount);

  // Read as String *****************************
  // var lChunk := TEncoding.UTF8.GetString(TBytes(AChunk), 0, AChunkLength);

  // Use bytes directly *************************
  var lFileName := 'chunk_' + TInterlocked.Add(gInc, 1).ToString + '.data';
  var lFile := TFile.Open(lFileName, TFileMode.fmCreate);
  try
    lFile.Write(TBytes(AChunk), AChunkLength);
  finally
    lFile.Free;
  end;
end;

procedure TMainForm.NetHTTPClient1ValidateServerCertificate(const Sender: TObject; const ARequest: TURLRequest;
  const Certificate: TCertificate; var Accepted: Boolean);
begin
  Accepted := True;
end;

procedure TMainForm.NetHTTPClientClick(Sender: TObject);
begin
  LogClear;
  Log('** NetHTTPClient **');
  var lResp := NetHTTPClient1.Get('https://www.danieleteti.it');
  Log(lResp.ContentAsString());
end;

procedure TMainForm.OnValidateCertificateEvent(const Sender: TObject; const ARequest: TURLRequest;
  const Certificate: TCertificate; var Accepted: Boolean);
begin
  Log('OnValidateCertificateEvent');
  Log('Certificate.CertName', Certificate.CertName);
  Log('Certificate.SerialNum', Certificate.SerialNum);
  Log('Certificate.Expiry', DateToStr(Certificate.Expiry));
  Log('Certificate.Start', DateToStr(Certificate.Start));
  Log('Certificate.Subject', Certificate.Subject);
  Log('Certificate.Issuer', Certificate.Issuer);
  Log('Certificate.AlgSignature', Certificate.AlgSignature);
  Log('Certificate.AlgEncryption', Certificate.AlgEncryption);
  Log('Certificate.KeySize', Certificate.KeySize);
  Accepted := True;
end;

procedure TMainForm.PrintSet(const VarName: String; const HashSet: THashSet<String>);
begin
  Log(VarName, String.Join(',', HashSet.ToArray));
end;

procedure TMainForm.Log(const Value: String; const BoolValue: Boolean);
begin
  Log(Value + ' = ' + BoolToStr(BoolValue, True));
end;

{ TMyClass }

constructor TMyClass.Create(Value: String);
begin
  inherited Create;
  fValue := Value;
end;

function TMyClass.ToString: string;
begin
  Result := fValue;
end;

procedure TMainForm.Log(const Value: String; const IntValues: TArray<Integer>);
begin
  Log(Value, '[' + TArray.ToString<Integer>(IntValues, ',') + ']');
end;

procedure TMainForm.Log(const Value: String; const Int64Value: UInt64);
begin
  Log(Value + ' = ' + Int64Value.ToString);
end;

end.
